import time
from snappi_ixnetwork.timer import Timer
from snappi_ixnetwork.customfield import CustomField


class TrafficItem(CustomField):
    """TrafficItem configuration

    Args
    ----
    - ixnetworkapi (Api): instance of the ixnetworkapi class

    Note: Sometime IxNetwork field type not matches with model packet field type
    Please use '@' to specify model packet field type. Ex
        'ether_type': 'pfcPause.header.header.ethertype@int'
    """

    _RESULT_COLUMNS = [
        ("frames_tx", "Tx Frames", int),
        ("frames_rx", "Rx Frames", int),
        ("frames_tx_rate", "Tx Frame Rate", float),
        ("frames_rx_rate", "Rx Frame Rate", float),
        ("bytes_tx", "Tx Bytes", int),
        ("bytes_rx", "Rx Bytes", int),
        ("loss", "Loss %", float),
        # ('bytes_tx_rate', 'Tx Rate (Bps)', float),
        # ('bytes_rx_rate', 'Rx Rate (Bps)', float),
    ]

    _RESULT_LATENCY_STORE_FORWARD = [
        ("minimum_ns", "Store-Forward Min Latency (ns)", int),
        ("maximum_ns", "Store-Forward Max Latency (ns)", int),
        ("average_ns", "Store-Forward Avg Latency (ns)", int),
    ]

    _RESULT_LATENCY_CUT_THROUGH = [
        ("minimum_ns", "Cut-Through Min Latency (ns)", int),
        ("maximum_ns", "Cut-Through Max Latency (ns)", int),
        ("average_ns", "Cut-Through Avg Latency (ns)", int),
    ]

    _RESULT_TIMESTAMP = [
        ("first_timestamp_ns", "First TimeStamp", int),
        ("last_timestamp_ns", "Last TimeStamp", int),
    ]

    _STACK_IGNORE = ["ethernet.fcs", "pfcPause.fcs"]

    _TYPE_TO_HEADER = {
        "ethernet": "ethernet",
        "pfcPause": "pfcpause",
        "vlan": "vlan",
        "ipv4": "ipv4",
        "ipv6": "ipv6",
        "tcp": "tcp",
        "udp": "udp",
        "gtpu": "gtpv1",
        "gTPuOptionalFields": "gtpv1option",
        "custom": "custom",
    }

    _HEADER_TO_TYPE = {
        "ethernet": "ethernet",
        "pfcpause": "pfcPause",
        "vlan": "vlan",
        "ipv4": "ipv4",
        "ipv6": "ipv6",
        "tcp": "tcp",
        "udp": "udp",
        "gtpv1": "gtpu",
        "gtpv1option": "gTPuOptionalFields",
        "custom": "custom",
    }

    _BIT_RATE_UNITS_TYPE = {
        "bps": "bitsPerSec",
        "kbps": "kbitsPerSec",
        "mbps": "mbitsPerSec",
        "gbps": "mbytesPerSec",
    }

    _LATENCY = {"cut_through": "cutThrough", "store_forward": "storeForward"}

    _PFCPAUSE = {
        "dst": "pfcPause.header.header.dstAddress",
        "src": "pfcPause.header.header.srcAddress",
        "ether_type": "pfcPause.header.header.ethertype@int",
        "control_op_code": "pfcPause.header.macControl.controlOpcode@int",
        "class_enable_vector": "pfcPause.header.macControl.priorityEnableVector@int",
        "pause_class_0": "pfcPause.header.macControl.pauseQuanta.pfcQueue0@int",
        "pause_class_1": "pfcPause.header.macControl.pauseQuanta.pfcQueue1@int",
        "pause_class_2": "pfcPause.header.macControl.pauseQuanta.pfcQueue2@int",
        "pause_class_3": "pfcPause.header.macControl.pauseQuanta.pfcQueue3@int",
        "pause_class_4": "pfcPause.header.macControl.pauseQuanta.pfcQueue4@int",
        "pause_class_5": "pfcPause.header.macControl.pauseQuanta.pfcQueue5@int",
        "pause_class_6": "pfcPause.header.macControl.pauseQuanta.pfcQueue6@int",
        "pause_class_7": "pfcPause.header.macControl.pauseQuanta.pfcQueue7@int",
    }

    _ETHERNET = {
        "dst": "ethernet.header.destinationAddress",
        "src": "ethernet.header.sourceAddress",
        "ether_type": "ethernet.header.etherType@int",
        "pfc_queue": "ethernet.header.pfcQueue",
    }

    _VLAN = {
        "id": "vlan.header.vlanTag.vlanID",
        "cfi": "vlan.header.vlanTag.cfi",
        "priority": "vlan.header.vlanTag.vlanUserPriority",
        "protocol": "vlan.header.protocolID",
    }

    _IPV4 = {
        "version": "ipv4.header.version",
        "header_length": "ipv4.header.headerLength",
        "priority": "_ipv4_priority",
        "total_length": "ipv4.header.totalLength",
        "identification": "ipv4.header.identification",
        "reserved": "ipv4.header.flags.reserved",
        "dont_fragment": "ipv4.header.flags.fragment",
        "more_fragments": "ipv4.header.flags.lastFragment",
        "fragment_offset": "ipv4.header.fragmentOffset",
        "time_to_live": "ipv4.header.ttl",
        "protocol": "ipv4.header.protocol",
        "header_checksum": "ipv4.header.checksum",
        "src": "ipv4.header.srcIp",
        "dst": "ipv4.header.dstIp",
    }

    _IPV6 = {
        "version": "ipv6.header.versionTrafficClassFlowLabel.version",
        "traffic_class": "ipv6.header.versionTrafficClassFlowLabel.trafficClass",
        "flow_label": "ipv6.header.versionTrafficClassFlowLabel.flowLabel",
        "payload_length": "ipv6.header.payloadLength",
        "next_header": "ipv6.header.nextHeader",
        "hop_limit": "ipv6.header.hopLimit",
        "src": "ipv6.header.srcIP",
        "dst": "ipv6.header.dstIP",
    }

    _TOS = {
        "precedence": "ipv4.header.priority.tos.precedence",
        "delay": "ipv4.header.priority.tos.delay",
        "throughput": "ipv4.header.priority.tos.throughput",
        "reliability": "ipv4.header.priority.tos.reliability",
        "monetary": "ipv4.header.priority.tos.monetary",
        "unused": "ipv4.header.priority.tos.unused",
    }

    _TCP = {
        "src_port": "tcp.header.srcPort",
        "dst_port": "tcp.header.dstPort",
        "ecn_ns": "tcp.header.ecn.nsBit",
        "ecn_cwr": "tcp.header.ecn.cwrBit",
        "ecn_echo": "tcp.header.ecn.ecnEchoBit",
        "ctl_urg": "tcp.header.controlBits.urgBit",
        "ctl_ack": "tcp.header.controlBits.ackBit",
        "ctl_psh": "tcp.header.controlBits.pshBit",
        "ctl_rst": "tcp.header.controlBits.rstBit",
        "ctl_syn": "tcp.header.controlBits.synBit",
        "ctl_fin": "tcp.header.controlBits.finBit",
    }

    _UDP = {
        "src_port": "udp.header.srcPort",
        "dst_port": "udp.header.dstPort",
        "length": "udp.header.length",
        "checksum": "udp.header.checksum",
    }

    _GTPV1 = {
        "version": "gtpu.header.version",
        "protocol_type": "gtpu.header.pt",
        "reserved": "gtpu.header.reserved",
        "e_flag": "gtpu.header.e",
        "s_flag": "gtpu.header.s",
        "pn_flag": "gtpu.header.n",
        "message_type": "tpu.header.type",
        "message_length": "gtpu.header.totalLength",
        "teid": "gtpu.header.teid",
    }

    _GTPV1OPTION = {
        "squence_number": "gTPuOptionalFields.header.sequenceNumber",
        "n_pdu_number": "gTPuOptionalFields.header.npduNumber",
        "next_extension_header_type": "gTPuOptionalFields.header.nextExtHdrField",
    }

    _CUSTOM = "_custom_headers"

    def __init__(self, ixnetworkapi):
        self._api = ixnetworkapi
        self.has_latency = False
        self._flow_timeout = 10

    def config(self):
        """Configure config.flows onto Ixnetwork.Traffic.TrafficItem

        CRUD
        ----
        - DELETE any TrafficItem.Name that does not exist in config.flows
        - CREATE TrafficItem for any config.flows[*].name that does not exist
        - UPDATE TrafficItem for any config.flows[*].name that exists
        """
        ixn_traffic_item = self._api._traffic_item
        self.flows_has_latency = []
        self.flows_has_timestamp = []
        self.flows_has_loss = []
        self.latency_mode = None
        self._api._remove(ixn_traffic_item, self._api.snappi_config.flows)
        if len(self._api.snappi_config.flows) > 0:
            for flow in self._api.snappi_config.flows:
                self._endpoint_validation(flow)
                args = {
                    "Name": flow.name,
                    "TrafficItemType": "l2L3",
                    "TrafficType": self._get_traffic_type(flow),
                    "SrcDestMesh": self._get_mesh_type(flow),
                }
                ixn_traffic_item.find(
                    Name="^%s$" % self._api.special_char(flow.name),
                    TrafficType=args["TrafficType"],
                )
                if len(ixn_traffic_item) == 0:
                    ixn_traffic_item.add(**args)
                else:
                    self._update(ixn_traffic_item, **args)
                self._configure_endpoint(
                    ixn_traffic_item.EndpointSet, flow.tx_rx
                )
                metrics = flow.get("metrics")
                if metrics is not None and metrics.enable is True:
                    self._configure_tracking(flow, ixn_traffic_item.Tracking)
                    latency = metrics.get("latency")
                    if latency is not None and latency.enable is True:
                        self.flows_has_latency.append(flow.name)
                        self._process_latency(latency)
                    timestamps = metrics.get("timestamps")
                    if timestamps is True:
                        self.flows_has_timestamp.append(flow.name)
                    loss = metrics.get("loss")
                    if loss is True:
                        self.flows_has_loss.append(flow.name)
                ixn_ce = ixn_traffic_item.ConfigElement.find()
                hl_stream_count = len(ixn_traffic_item.HighLevelStream.find())
                self._configure_stack(
                    ixn_ce, flow.get("packet", with_default=True)
                )
                self._configure_size(
                    ixn_ce, flow.get("size", with_default=True)
                )
                self._configure_rate(
                    ixn_ce, flow.get("rate", with_default=True)
                )
                self._configure_tx_control(
                    ixn_ce, hl_stream_count, flow.duration
                )
            self._configure_options()
            self._configure_latency()

    def _process_latency(self, latency):
        if self.latency_mode is None:
            if latency.mode is not None:
                self.latency_mode = latency.mode
            else:
                self.latency_mode = "store_forward"
        else:
            if latency.mode is not None and self.latency_mode != latency.mode:
                raise Exception("Latency mode needs to be same for all flows")

    def _configure_latency(self):
        ixn_latency = self._api._traffic.Statistics.Latency
        if self.latency_mode is not None:
            self.has_latency = True
            ixn_CpdpConvergence = self._api._traffic.Statistics.CpdpConvergence
            ixn_CpdpConvergence.Enabled = False
            ixn_latency.Enabled = True
            ixn_latency.Mode = TrafficItem._LATENCY[self.latency_mode]
        else:
            self.has_latency = False
            ixn_latency.Enabled = False

    def _configure_tracking(self, flow, ixn_tracking):
        """Set tracking options"""
        ixn_tracking.find()
        tracking_options = ["trackingenabled0"]
        if flow.tx_rx.choice == "device":
            tracking_options.append("sourceDestPortPair0")
        if set(tracking_options) != set(ixn_tracking.TrackBy):
            ixn_tracking.update(TrackBy=tracking_options)

    def _configure_options(self):
        enable_min_frame_size = False
        for flow in self._api.snappi_config.flows:
            if (
                len(flow.packet) == 1
                and flow.packet[0].parent.choice == "pfcpause"
            ):
                enable_min_frame_size = True
                break
        if self._api._traffic.EnableMinFrameSize != enable_min_frame_size:
            self._api._traffic.EnableMinFrameSize = enable_min_frame_size

    def _endpoint_validation(self, flow):
        if flow.tx_rx.choice is None:
            raise ValueError(
                "%s Flow.tx_rx property cannot be None" % flow.name
            )
        if flow.tx_rx.choice == "device":
            device = flow.tx_rx.device
            if not isinstance(device.tx_names, list) or not isinstance(
                device.rx_names, list
            ):
                raise ValueError(
                    "device tx_names and rx_names must be a list "
                    "in flow %s" % flow.name
                )
            if len(device.tx_names) != len(set(device.tx_names)):
                raise ValueError(
                    "All names in device tx_names "
                    "must be unique for flow %s" % flow.name
                )
            if len(device.rx_names) != len(set(device.rx_names)):
                raise ValueError(
                    "All names in device rx_names "
                    "must be unique for flow %s" % flow.name
                )

    def _get_mesh_type(self, flow):
        if flow.tx_rx.choice == "port":
            mesh_type = "oneToOne"
        else:
            device = flow.tx_rx.device
            if device.mode == "mesh" or device.mode is None:
                mesh_type = "manyToMany"
            else:
                mesh_type = "oneToOne"
                if len(device.tx_names) != len(device.rx_names):
                    raise ValueError(
                        "Length of device tx_names and rx_names "
                        "must be same for device mode ONE_TO_ONE in flow %s"
                        % flow.name
                    )
        return mesh_type

    def _get_traffic_type(self, flow):
        if flow.tx_rx.choice == "port":
            encap = "raw"
        else:
            encap_list = []
            device = flow.tx_rx.device
            if device.tx_names is not None:
                for tx_name in device.tx_names:
                    encap_list.append(self._api.get_device_encap(tx_name))
            if device.rx_names is not None:
                for rx_name in device.rx_names:
                    encap_list.append(self._api.get_device_encap(rx_name))
            if len(set(encap_list)) == 1:
                encap = encap_list[0]
            else:
                raise Exception(
                    "All devices identified in tx_names and rx_names "
                    "must be of same type in flow %s" % flow.name
                )
        return encap

    def _configure_endpoint(self, ixn_endpoint_set, endpoint):
        """Transform flow.tx_rx to /trafficItem/endpointSet
        The model allows for only one endpointSet per traffic item
        """
        args = {"Sources": [], "Destinations": []}
        if endpoint.choice == "port":
            args["Sources"].append(
                self._api.get_ixn_object(endpoint.port.tx_name)
                .Protocols.find()
                .href
            )
            if endpoint.port.rx_name != None:
                args["Destinations"].append(
                    self._api.get_ixn_object(endpoint.port.rx_name)
                    .Protocols.find()
                    .href
                )
        else:
            for device_name in endpoint.device.tx_names:
                args["Sources"].append(self._api.get_ixn_href(device_name))
            for device_name in endpoint.device.rx_names:
                args["Destinations"].append(
                    self._api.get_ixn_href(device_name)
                )
        ixn_endpoint_set.find()
        if len(ixn_endpoint_set) > 1:
            ixn_endpoint_set.remove()
        if len(ixn_endpoint_set) == 0:
            ixn_endpoint_set.add(**args)
        elif (
            ixn_endpoint_set.Sources != args["Sources"]
            or ixn_endpoint_set.Destinations != args["Destinations"]
            or len(ixn_endpoint_set.parent.ConfigElement.find()) == 0
        ):
            self._update(ixn_endpoint_set, **args)

    def _update(self, ixn_object, **kwargs):
        from ixnetwork_restpy.base import Base

        update = False
        for name, value in kwargs.items():
            if (
                isinstance(value, list)
                and len(value) > 0
                and isinstance(value[0], Base)
            ):
                value = [item.href for item in value]
            elif isinstance(value, Base):
                value = value.href
            if getattr(ixn_object, name) != value:
                update = True
        if update is True:
            ixn_object.update(**kwargs)

    def _configure_stack(self, ixn_stream, headers):
        """Transform flow.packets[0..n] to /traffic/trafficItem/configElement/stack
        The len of the headers list is the definitive list which means add/remove
        any stack items so that the stack list matches the headers list.
        If the headers list is empty then use the traffic generator default stack.
        """
        headers = self.adjust_header(headers)
        ixn_stack = ixn_stream.Stack.find()
        for i in range(0, len(headers)):
            header = headers[i]
            if len(ixn_stack) <= i:
                stack = self._add_stack(ixn_stream, stack, header)
            else:
                stack_type_id = ixn_stack[i].StackTypeId
                if stack_type_id in self._STACK_IGNORE:
                    stack = self._add_stack(ixn_stream, stack, header)
                elif stack_type_id not in TrafficItem._TYPE_TO_HEADER:
                    stack = self._add_stack(ixn_stream, ixn_stack[i], header)
                elif (
                    TrafficItem._TYPE_TO_HEADER[stack_type_id] != header.choice
                ):
                    stack = self._add_stack(ixn_stream, ixn_stack[i], header)
                else:
                    stack = ixn_stack[i]
            self._configure_field(stack.Field, header)

        # scan and compare new stack to overcome IxNetwork stack serialization
        # then remove additional stack
        if len(headers) == 0:
            return
        stacks_to_remove = []
        ixn_stack = ixn_stream.Stack.find()
        header_index = 0
        for i in range(0, len(ixn_stack)):
            stack_type_id = ixn_stack[i].StackTypeId
            if stack_type_id in self._STACK_IGNORE:
                continue
            if len(headers) <= header_index:
                stacks_to_remove.append(ixn_stack[i])
                continue
            if (
                TrafficItem._TYPE_TO_HEADER[stack_type_id]
                != headers[header_index].choice
            ):
                stacks_to_remove.append(ixn_stack[i])
            else:
                header_index += 1
        for stack in stacks_to_remove[::-1]:
            stack.Remove()

    def _add_stack(self, ixn_stream, ixn_stack, header):
        type_id = "^%s$" % TrafficItem._HEADER_TO_TYPE[header.choice]
        template = self._api._traffic.ProtocolTemplate.find(
            StackTypeId=type_id
        )
        stack_href = ixn_stack.AppendProtocol(template)
        return ixn_stream.Stack.read(stack_href)

    def _configure_field(self, ixn_field, header, field_choice=False):
        """Transform flow.packets[0..n].header.choice to /traffic/trafficItem/configElement/stack/field"""
        field_map = getattr(self, "_%s" % header.choice.upper())
        packet = getattr(header, header.choice)
        if isinstance(field_map, dict) is False:
            method = getattr(self, field_map)
            method(ixn_field, packet)
            return

        for packet_field_name in dir(packet):
            if packet_field_name in field_map:
                pattern = packet.get(packet_field_name, with_default=True)
                if pattern is not None:
                    field_type_id = field_map[packet_field_name]
                    self._configure_pattern(
                        ixn_field, field_type_id, pattern, field_choice
                    )

    def _configure_pattern(
        self, ixn_field, field_type_id, pattern, field_choice=False
    ):
        def get_value(field_value):
            if field_type is None:
                return field_value
            ixn_type = ixn_field.ValueFormat
            if field_type == "int" and ixn_type == "hex":
                field_value = str(hex(int(field_value)))[2:]
            return field_value

        custom_field = getattr(self, field_type_id, None)
        if custom_field is not None:
            if pattern.choice is not None:
                custom_field(ixn_field, pattern)
            return
        id_type = field_type_id.split("@")
        field_type_id, field_type = (
            id_type if len(id_type) > 1 else [id_type[0], None]
        )
        ixn_field = ixn_field.find(FieldTypeId=field_type_id)
        if pattern.choice is None:
            self._set_default(ixn_field, field_choice)
            return

        if pattern.choice == "value":
            ixn_field.update(
                Auto=False,
                ActiveFieldChoice=field_choice,
                ValueType="singleValue",
                SingleValue=get_value(pattern.value),
            )
        elif pattern.choice == "values":
            ixn_field.update(
                Auto=False,
                ActiveFieldChoice=field_choice,
                ValueType="valueList",
                ValueList=[get_value(v) for v in pattern.values],
            )
        elif pattern.choice == "increment":
            ixn_field.update(
                Auto=False,
                ValueType="increment",
                ActiveFieldChoice=field_choice,
                StartValue=get_value(pattern.increment.start),
                StepValue=pattern.increment.step,
                CountValue=pattern.increment.count,
            )
        elif pattern.choice == "decrement":
            ixn_field.update(
                Auto=False,
                ValueType="decrement",
                ActiveFieldChoice=field_choice,
                StartValue=get_value(pattern.decrement.start),
                StepValue=pattern.decrement.step,
                CountValue=pattern.decrement.count,
            )
        elif pattern.choice == "random":
            ixn_field.update(
                Auto=False,
                ActiveFieldChoice=field_choice,
                ValueType="repeatableRandomRange",
                MinValue=pattern.random.min,
                MaxValue=pattern.random.max,
                StepValue=pattern.random.step,
                Seed=pattern.random.seed,
                CountValue=pattern.random.count,
            )
        else:
            # TBD: add to set_config errors - invalid pattern specified
            pass

        if pattern.get("metric_group") is not None:
            ixn_field.TrackingEnabled = True
            self._api.ixn_objects[pattern.metric_group] = ixn_field.href

    def _set_default(self, ixn_field, field_choice):
        """We are setting all the field to default. Otherwise test is keeping the same value from previous run."""
        if ixn_field.ReadOnly:
            return

        if ixn_field.SupportsAuto:
            if ixn_field.Auto is not True:
                ixn_field.Auto = True
        else:
            ixn_field.update(
                Auto=False,
                ActiveFieldChoice=field_choice,
                ValueType="singleValue",
                SingleValue=ixn_field.DefaultValue,
            )

    def _configure_size(self, ixn_stream, size):
        """Transform frameSize flows.size to /traffic/trafficItem[*]/configElement[*]/frameSize"""
        if size.choice is None:
            return
        ixn_frame_size = ixn_stream.FrameSize
        args = {}
        if size.choice == "fixed":
            args["Type"] = "fixed"
            args["FixedSize"] = size.fixed
        elif size.choice == "increment":
            args["Type"] = "increment"
            args["IncrementFrom"] = size.increment.start
            args["IncrementTo"] = size.increment.end
            args["IncrementStep"] = size.increment.step
        elif size.choice == "random":
            args["Type"] = "random"
            args["RandomMin"] = size.random.min
            args["RandomMax"] = size.random.max
        else:
            print(
                "Warning - We need to implement this %s choice" % size.choice
            )
        self._update(ixn_frame_size, **args)

    def _configure_rate(self, ixn_stream, rate):
        """Transform frameRate flows.rate to /traffic/trafficItem[*]/configElement[*]/frameRate"""
        if rate.choice is None:
            return
        ixn_frame_rate = ixn_stream.FrameRate
        args = {}
        value = None
        if rate.choice == "percentage":
            args["Type"] = "percentLineRate"
            value = rate.percentage
        elif rate.choice == "pps":
            args["Type"] = "framesPerSecond"
            value = rate.pps
        else:
            args["Type"] = "bitsPerSecond"
            args["BitRateUnitsType"] = TrafficItem._BIT_RATE_UNITS_TYPE[
                rate.choice
            ]
            value = getattr(rate, rate.choice)
        args["Rate"] = value
        self._update(ixn_frame_rate, **args)

    def _configure_delay(self, parent, args):
        delay = parent.get("delay", with_default=True)
        if delay.choice is not None:
            value = getattr(delay, delay.choice, None)
            if value is None:
                raise Exception("Delay must be of type <int>")
            if isinstance(value, float) and not float.is_integer(value):
                self._api.warning(
                    "Cast Delay to <int> due to software limitation"
                )
            if delay.choice == "microseconds":
                args["StartDelayUnits"] = "nanoseconds"
                args["StartDelay"] = value * 1000
            else:
                args["StartDelayUnits"] = delay.choice
                args["StartDelay"] = value

    def _configure_tx_control(self, ixn_stream, hl_stream_count, duration):
        """Transform duration flows.duration to /traffic/trafficItem[*]/configElement[*]/TransmissionControl"""
        if duration.choice is None:
            return
        ixn_tx_control = ixn_stream.TransmissionControl
        args = {}
        if duration.choice == "continuous":
            args["Type"] = "continuous"
            args["MinGapBytes"] = duration.continuous.gap
            self._configure_delay(duration.continuous, args)
        elif duration.choice == "fixed_packets":
            args["Type"] = "fixedFrameCount"
            args["FrameCount"] = (
                duration.fixed_packets.packets / hl_stream_count
            )
            args["MinGapBytes"] = duration.fixed_packets.gap
            self._configure_delay(duration.fixed_packets, args)
        elif duration.choice == "fixed_seconds":
            args["Type"] = "fixedDuration"
            args["Duration"] = duration.fixed_seconds.seconds
            args["MinGapBytes"] = duration.fixed_seconds.gap
            self._configure_delay(duration.fixed_seconds, args)
        elif duration.choice == "burst":
            if (
                duration.burst.bursts is not None
                and int(duration.burst.bursts) > 0
            ):
                args["Type"] = "burstFixedDuration"
                args["RepeatBurst"] = duration.burst.bursts
            else:
                args["Type"] = "custom"
            args["BurstPacketCount"] = duration.burst.packets
            args["MinGapBytes"] = duration.burst.gap
            args["EnableInterBurstGap"] = True
            inter_burst_gap = duration.burst.get(
                "inter_burst_gap", with_default=True
            )
            if inter_burst_gap.choice is not None:
                value = getattr(inter_burst_gap, inter_burst_gap.choice, None)
                if value is None:
                    raise Exception(
                        "Inter packet gap mush be of type some <int> value"
                    )
                if inter_burst_gap.choice == "microseconds":
                    args["InterBurstGap"] = value * 1000
                    args["InterBurstGapUnits"] = "nanoseconds"
                else:
                    args["InterBurstGap"] = value
                    args["InterBurstGapUnits"] = inter_burst_gap.choice
        self._update(ixn_tx_control, **args)

    def transmit(self, request):
        """Set flow transmit
        1) If start then start any device protocols that are traffic dependent
        2) If start then generate and apply traffic
        3) Execute requested transmit action (start|stop|pause|resume)
        """
        regex = ""
        flow_names = [flow.name for flow in self._api._config.flows]
        if request and request.flow_names:
            flow_names = request.flow_names
        if len(flow_names) == 1:
            regex = "^%s$" % self._api.special_char(flow_names)[0]
        elif len(flow_names) > 1:
            regex = "^(%s)$" % "|".join(self._api.special_char(flow_names))

        if request.state == "start":
            if len(self._api._topology.find()) > 0:
                with Timer(self._api, "Devices start"):
                    self._api._ixnetwork.StartAllProtocols("sync")
                    self._api.check_protocol_statistics()
            if len(self._api._traffic_item.find()) == 0:
                return
            self._api._traffic_item.find(State="^unapplied$")
            if len(self._api._traffic_item) > 0:
                with Timer(self._api, "Flows generate/apply"):
                    self._api._traffic_item.Generate()
                    self._api._traffic.Apply()
            self._api._traffic_item.find(State="^started$")
            if len(self._api._traffic_item) == 0:
                with Timer(self._api, "Flows clear statistics"):
                    self._api._ixnetwork.ClearStats(
                        [
                            "waitForPortStatsRefresh",
                            "waitForTrafficStatsRefresh",
                        ]
                    )
            self._api.capture._start_capture()
        self._api._traffic_item.find(Name=regex)
        if len(self._api._traffic_item) > 0:
            if request.state == "start":
                self._api._traffic_item.find(
                    Name=regex, Suspend=True, State="^started$"
                )
                if len(self._api._traffic_item) > 0:
                    with Timer(self._api, "Flows resume"):
                        self._api._traffic_item.PauseStatelessTraffic(False)
                self._api._traffic_item.find(Name=regex, State="^stopped$")
                if len(self._api._traffic_item) > 0:
                    with Timer(self._api, "Flows start"):
                        self._api._traffic_item.StartStatelessTrafficBlocking()
            elif request.state == "stop":
                self._api._traffic_item.find(Name=regex, State="^started$")
                if len(self._api._traffic_item) > 0:
                    with Timer(self._api, "Flows stop"):
                        self._api._traffic_item.StopStatelessTrafficBlocking()
            elif request.state == "pause":
                self._api._traffic_item.find(Name=regex, State="^started$")
                if len(self._api._traffic_item) > 0:
                    with Timer(self._api, "Flows pause"):
                        self._api._traffic_item.PauseStatelessTraffic(True)
        if request.state == "stop":
            if len(self._api._topology.find()) > 0:
                with Timer(self._api, "Devices stop"):
                    self._api._ixnetwork.StopAllProtocols("sync")

    def _set_result_value(
        self, row, column_name, column_value, column_type=str
    ):
        if (
            len(self._column_names) > 0
            and column_name not in self._column_names
        ):
            return
        try:
            row[column_name] = column_type(column_value)
        except:
            if column_type.__name__ in ["float", "int"]:
                row[column_name] = 0
            else:
                row[column_type] = column_value

    def _get_state(self, state):
        """IxNetwork traffic item states
        error, locked, started,
        startedWaitingForStats, startedWaitingForStreams, stopped,
        stoppedWaitingForStats, txStopWatchExpected, unapplied
        """
        started_states = [
            "txStopWatchExpected",
            "locked",
            "started",
            "startedWaitingForStats",
            "startedWaitingForStreams",
            "stoppedWaitingForStats",
        ]
        if state in started_states:
            return "started"
        else:
            return "stopped"

    def results(self, request):
        """Return flow results"""
        # setup parameters
        self._column_names = request.get("metric_names")
        if self._column_names is None:
            self._column_names = []
        elif not isinstance(self._column_names, list):
            msg = "Invalid format of column_names passed {},\
                    expected list".format(
                self._column_names
            )
            raise Exception(msg)

        flow_names = request.get("flow_names")
        has_request_flow = True
        if flow_names is None or len(flow_names) == 0:
            has_request_flow = False
            flow_names = [flow.name for flow in self._api._config.flows]
        elif not isinstance(flow_names, list):
            msg = "Invalid format of flow_names passed {},\
                    expected list".format(
                flow_names
            )
            raise Exception(msg)
        final_flow_names = []
        for flow in self._api._config.flows:
            metrics = flow.get("metrics")
            if metrics is None:
                continue
            if metrics.enable is True and flow.name in flow_names:
                final_flow_names.append(flow.name)
        if len(final_flow_names) == 0:
            msg = """
            To fetch flow metrics at least one flow shall have metric enabled
            """
            raise Exception(msg.strip())
        diff = set(flow_names).difference(final_flow_names)
        if len(diff) > 0 and has_request_flow is True:
            raise Exception(
                "Flow metrics is not enabled on flows {}".format(
                    "".join(list(diff))
                )
            )
        flow_names = final_flow_names
        regfilter = {"property": "name", "regex": ".*"}
        regfilter["regex"] = "^(%s)$" % "|".join(
            self._api.special_char(flow_names)
        )

        flow_count = len(flow_names)
        ixn_page = self._api._ixnetwork.Statistics.View.find(
            Caption="Flow Statistics"
        ).Page
        if ixn_page.PageSize < flow_count:
            ixn_page.PageSize = flow_count

        # initialize result values
        flow_rows = {}
        for traffic_item in self._api.select_traffic_items(
            traffic_item_filters=[regfilter]
        ).values():
            for stream in traffic_item["highLevelStream"]:
                for rx_port_name in stream["rxPortNames"]:
                    flow_row = {}
                    self._set_result_value(
                        flow_row, "name", traffic_item["name"]
                    )
                    self._set_result_value(
                        flow_row,
                        "transmit",
                        self._get_state(traffic_item["state"]),
                    )
                    self._set_result_value(
                        flow_row, "port_tx", stream["txPortName"]
                    )
                    self._set_result_value(flow_row, "port_rx", rx_port_name)
                    # init all columns with corresponding zero-values so that
                    # the underlying dictionary contains all requested columns
                    # in an event of unwanted exceptions
                    for (
                        external_name,
                        _,
                        external_type,
                    ) in self._RESULT_COLUMNS:
                        self._set_result_value(
                            flow_row, external_name, 0, external_type
                        )
                    flow_rows[
                        traffic_item["name"]
                        + stream["txPortName"]
                        + rx_port_name
                    ] = flow_row

        flow_stat = self._api.assistant.StatViewAssistant("Flow Statistics")
        for row in flow_stat.Rows:
            name = row["Traffic Item"]
            if len(flow_names) > 0 and name not in flow_names:
                continue
            tx_port = row["Tx Port"]
            rx_port = row["Rx Port"]
            if name + tx_port + rx_port in flow_rows:
                flow_row = flow_rows[name + tx_port + rx_port]
                if (
                    float(row["Tx Frame Rate"]) > 0
                    or int(row["Tx Frames"]) == 0
                ):
                    flow_row["transmit"] = "started"
                else:
                    flow_row["transmit"] = "stopped"
                for (
                    external_name,
                    internal_name,
                    external_type,
                ) in self._RESULT_COLUMNS:
                    # keep plugging values for next columns even if the
                    # current one raises exception
                    try:
                        self._set_result_value(
                            flow_row,
                            external_name,
                            row[internal_name],
                            external_type,
                        )
                    except Exception:
                        # TODO print a warning maybe ?
                        pass
                if name in self.flows_has_latency:
                    self._construct_latency(flow_row, row)
                if name in self.flows_has_timestamp:
                    self._construct_timestamp(flow_row, row)
                if name not in self.flows_has_loss:
                    flow_row.pop("loss")
        return list(flow_rows.values())

    def _construct_latency(self, flow_row, row):
        if self.latency_mode == "store_forward":
            latency_map = TrafficItem._RESULT_LATENCY_STORE_FORWARD
        else:
            latency_map = TrafficItem._RESULT_LATENCY_CUT_THROUGH
        latency_result = {}
        for external_name, internal_name, external_type in latency_map:
            if internal_name not in row.Columns:
                raise Exception(
                    "Could not fetch column %s in latency metrics"
                    % internal_name
                )
            try:
                self._set_result_value(
                    latency_result,
                    external_name,
                    row[internal_name],
                    external_type,
                )
            except Exception as exception_err:
                self._api.warning(
                    "Could not fetch latency metrics: %s" % exception_err
                )
                pass
        if len(latency_result) > 0:
            flow_row["latency"] = latency_result

    def _construct_timestamp(self, flow_row, row):
        timestamp_result = {}
        for (
            external_name,
            internal_name,
            external_type,
        ) in TrafficItem._RESULT_TIMESTAMP:
            if internal_name not in row.Columns:
                raise Exception(
                    "Could not fetch column %s in timestamp metrics"
                    % internal_name
                )

            try:
                val = row[internal_name].split(":")
                mul = [3600, 60, 1]
                sv = sum(
                    [
                        int(float(v) * 10 ** 9 + 0.1) * mul[i]
                        for i, v in enumerate(val)
                    ]
                )
                self._set_result_value(
                    timestamp_result, external_name, sv, external_type
                )
            except Exception as exception_err:
                self._api.warning(
                    "Could not fetch timestamp metrics: %s" % exception_err
                )
                pass
        if len(timestamp_result) > 0:
            flow_row["timestamps"] = timestamp_result



import re
from collections import namedtuple


class ConfigureBgp(object):
    _BGP_AS_SET_MODE = {
        "do_not_include_as": "dontincludelocalas",
        "include_as_seq": "includelocalasasasseq",
        "include_as_set": "includelocalasasasset",
        "include_as_seq_confed": "includelocalasasasseqconfederation",
        "include_as_set_confed": "includelocalasasassetconfederation",
        "prepend_as_to_first_segment": "prependlocalastofirstsegment",
    }

    _BGP_AS_MODE = {
        "do_not_include_local_as": "dontincludelocalas",
        "include_as_seq": "includelocalasasasseq",
        "include_as_set": "includelocalasasasset",
        "include_as_confed_seq": "includelocalasasasseqconfederation",
        "include_as_confed_set": "includelocalasasassetconfederation",
        "prepend_to_first_segment": "prependlocalastofirstsegment",
    }

    _BGP_SEG_TYPE = {
        "as_seq": "asseq",
        "as_set": "asset",
        "as_confed_seq": "asseqconfederation",
        "as_confed_set": "assetconfederation",
    }

    _BGP_COMMUNITY_TYPE = {
        "manual_as_number": "manual",
        "no_export": "noexport",
        "no_advertised": "noadvertised",
        "no_export_subconfed": "noexport_subconfed",
        "llgr_stale": "llgr_stale",
        "no_llgr": "no_llgr",
    }

    _BGP_SR_TE = {
        "policy_type": {"ixn_attr": "policyType", "default": "ipv4"},
        "distinguisher": {"ixn_attr": "distinguisher", "default": "1"},
        "color": {"ixn_attr": "policyColor", "default": "101"},
        "ipv4_endpoint": {"ixn_attr": "endPointV4", "default": "0.0.0.0"},
        "ipv6_endpoint": {
            "ixn_attr": "endPointV6",
            "default": "0:0:0:0:0:0:0:0",
        },
    }

    _SRTE_NEXT_HOP = {
        "next_hop_mode": {
            "ixn_attr": "setNextHop",
            "default": "sameaslocalip",
            "enum_map": {"local_ip": "sameaslocalip", "manual": "manually"},
        },
        "next_hop_address_type": {
            "ixn_attr": "setNextHopIpType",
            "default": "ipv4",
        },
        "ipv4_address": {"ixn_attr": "ipv4NextHop", "default": "0.0.0.0"},
        "ipv6_address": {"ixn_attr": "ipv6NextHop", "default": "::"},
    }

    _SRTE_ADDPATH = {"path_id": {"ixn_attr": "addPathId", "default": "1"}}

    _SRTE_AS_PATH = {
        "override_peer_as_set_mode": {
            "ixn_attr": "overridePeerAsSetMode",
            "default": False,
        },
        "as_set_mode": {
            "ixn_attr": "asSetMode",
            "default": "includelocalasasasseq",
            "enum_map": _BGP_AS_MODE,
        },
    }

    _SRTE_ASPATH_SEGMENT = {
        "segment_type": {
            "ixn_attr": "segmentType",
            "default": "asset",
            "enum_map": _BGP_SEG_TYPE,
        }
    }

    _REMOTE_ENDPOINT_SUB_TLV = {
        "as_number": {"ixn_attr": "as4Number", "default": "0"},
        "address_family": {"ixn_attr": "addressFamily", "default": "ipv4"},
        "ipv4_address": {
            "ixn_attr": "remoteEndpointIPv4",
            "default": "0.0.0.0",
        },
        "ipv6_address": {"ixn_attr": "remoteEndpointIPv6", "default": "::"},
    }

    _PREFERENCE_SUB_TLV = {
        "preference": {"ixn_attr": "prefValue", "default": "0"}
    }

    _BINDING_SUB_TLV = {
        "binding_sid_type": {
            "ixn_attr": "bindingSIDType",
            "default": "nobinding",
            "enum_map": {
                "no_binding": "nobinding",
                "four_octet_sid": "sid4",
                "ipv6_sid": "ipv6sid",
            },
        },
        "four_octet_sid": {"ixn_attr": "SID4Octet", "default": "0"},
        "bsid_as_mpls_label": {"ixn_attr": "useAsMPLSLabel", "default": False},
        "ipv6_sid": {"ixn_attr": "IPv6SID", "default": "::"},
        "s_flag": {"ixn_attr": "sflag", "default": False},
        "i_flag": {"ixn_attr": "iflag", "default": False},
        "remaining_flag_bits": {
            "ixn_attr": "remainingBits",
            "default": "0x01",
        },
    }

    _ENLP_SUB_TLV = {
        "explicit_null_label_policy": {
            "ixn_attr": "ENLPValue",
            "default": "4",
            "enum_map": {
                "reserved_enlp": "0",
                "push_ipv4_enlp": "1",
                "push_ipv6_enlp": "2",
                "push_ipv4_ipv6_enlp": "3",
                "do_not_push_enlp": "4",
            },
        }
    }

    _POLICIES_SEGMENT_LIST = {
        "segment_weight": {"ixn_attr": "weight", "default": "200"}
    }

    _SEGMENTS = {
        "segment_type": {
            "ixn_attr": "segmentType",
            "default": "mplssid",
            "enum_map": {"mpls_sid": "mplssid", "ipv6_sid": "ipv6sid"},
        },
        "mpls_label": {"ixn_attr": "label", "default": "16"},
        "mpls_tc": {"ixn_attr": "trafficClass", "default": "0"},
        "mpls_ttl": {"ixn_attr": "timeToLive", "default": "255"},
        "v_flag": {"ixn_attr": "vflag", "default": False},
        "remaining_flag_bits": {
            "ixn_attr": "remainingBits",
            "default": "0x01",
        },
        "ipv6_sid": {"ixn_attr": "ipv6SID", "default": "::"},
    }

    def __init__(self, ngpf):
        self._ngpf = ngpf
        self._api = ngpf._api
        self.update = ngpf.update
        self.configure_value = ngpf.configure_value
        self.get_xpath = ngpf.get_xpath
        self.select_node = ngpf.select_node
        self.select_child_node = ngpf.select_child_node

    def configure_bgpv4(self, ixn_parent, bgpv4, ixn_dg):
        ixn_bgpv4 = ixn_parent.BgpIpv4Peer
        self._api._remove(ixn_bgpv4, [bgpv4])
        name = self._api.special_char(bgpv4.name)
        args = {
            "Name": name,
        }
        ixn_bgpv4.find(Name="^%s$" % name)
        if len(ixn_bgpv4) == 0:
            ixn_bgpv4.add(**args)[-1]
        else:
            self.update(ixn_bgpv4, **args)
        self._api.ixn_objects[bgpv4.name] = ixn_bgpv4.href
        as_type = "internal"
        if bgpv4.get("as_type") is not None and bgpv4.get("as_type") == "ebgp":
            as_type = "external"
        bgp_xpath = self.get_xpath(ixn_bgpv4.href)
        self.configure_value(bgp_xpath, "type", as_type)
        as_bytes = bgpv4.get("as_number_width")
        if bgpv4.get("as_number") is not None:
            if as_bytes is None or as_bytes == "two":
                self.configure_value(
                    bgp_xpath, "localAs2Bytes", bgpv4.as_number
                )
            elif as_bytes == "four":
                self.configure_value(bgp_xpath, "enable4ByteAs", True)
                self.configure_value(
                    bgp_xpath, "localAs4Bytes", bgpv4.as_number
                )
            else:
                msg = "Please configure supported [two, four] as_number_width"
                raise Exception(msg)
        if bgpv4.get("dut_address") is not None:
            self.configure_value(bgp_xpath, "dutIp", bgpv4.dut_address)
        if bgpv4.get("as_number_set_mode"):
            self.configure_value(
                bgp_xpath,
                "asSetMode",
                bgpv4.as_number_set_mode,
                enum_map=ConfigureBgp._BGP_AS_SET_MODE,
            )
        # self._configure_pattern(ixn_dg.RouterData.RouterId, bgpv4.router_id)
        advanced = bgpv4.get("advanced")
        if advanced is not None:
            self.configure_value(
                bgp_xpath, "holdTimer", advanced.hold_time_interval
            )
            self.configure_value(
                bgp_xpath, "keepaliveTimer", advanced.keep_alive_interval
            )
            self.configure_value(bgp_xpath, "md5Key", advanced.md5_key)
            self.configure_value(
                bgp_xpath, "updateInterval", advanced.update_interval
            )
            self.configure_value(bgp_xpath, "ttl", advanced.time_to_live)
        if bgpv4.get("sr_te_policies") is not None:
            self._configure_sr_te(ixn_bgpv4, bgp_xpath, bgpv4.sr_te_policies)
        self._bgp_route_builder(ixn_dg, ixn_bgpv4, bgpv4)
        return ixn_bgpv4

    def _bgp_route_builder(self, ixn_dg, ixn_bgp, bgp):
        bgpv4_routes = bgp.bgpv4_routes
        bgpv6_routes = bgp.bgpv6_routes
        if len(bgpv4_routes) > 0:
            for route_range in bgpv4_routes:
                self._configure_bgpv4_route(ixn_dg, ixn_bgp, route_range)
        if len(bgpv6_routes) > 0:
            for route_range in bgpv6_routes:
                self._configure_bgpv6_route(ixn_dg, ixn_bgp, route_range)

    def _configure_bgpv4_route(self, ixn_dg, ixn_bgp, route_range):
        ixn_ng = ixn_dg.NetworkGroup
        name = self._api.special_char(route_range.name)
        args = {
            "Name": name,
        }
        ixn_ng.find(Name="^%s$" % name)
        if len(ixn_ng) == 0:
            self.stop_topology()
            ixn_ng.add(**args)[-1]
            ixn_pool = ixn_ng.Ipv4PrefixPools.add()
        else:
            self.update(ixn_ng, **args)
            ixn_pool = ixn_ng.Ipv4PrefixPools.find()
        ixn_pool.Connector.find().ConnectedTo = ixn_bgp.href
        if route_range.name is not None:
            self._api.ixn_objects[route_range.name] = ixn_ng.href
            self._api._device_encap[route_range.name] = "ipv4"
        pool_infos = self.select_node(
            ixn_pool.href,
            children=["bgpIPRouteProperty", "bgpV6IPRouteProperty"],
        )
        pool_xpath = pool_infos["xpath"]
        addresses = route_range.addresses
        if len(addresses) > 0:
            ixn_ng.Multiplier = len(addresses)
            route_addresses = RouteAddresses()
            for address in addresses:
                # below properties will set to default when
                # route_address is instantiated
                route_addresses.address = address.address
                route_addresses.step = address.step
                route_addresses.prefix = address.prefix
                route_addresses.count = address.count
            self.configure_value(
                pool_xpath, "networkAddress", route_addresses.address
            )
            self.configure_value(
                pool_xpath, "prefixAddrStep", route_addresses.step
            )
            self.configure_value(
                pool_xpath, "prefixLength", route_addresses.prefix
            )
            self.configure_value(
                pool_xpath, "numberOfAddressesAsy", route_addresses.count
            )
        if "bgpIPRouteProperty" in pool_infos:
            ixn_bgp_property = ixn_pool.BgpIPRouteProperty.find()
            property_xpath = pool_infos["bgpIPRouteProperty"][0]["xpath"]
        else:
            ixn_bgp_property = ixn_pool.BgpV6IPRouteProperty.find()
            property_xpath = pool_infos["bgpV6IPRouteProperty"][0]["xpath"]
        if route_range.get("next_hop_address"):
            self.configure_value(
                property_xpath, "ipv4NextHop", route_range.next_hop_address
            )
        if route_range.name is not None:
            ixn_bgp_property.Name = route_range.name
            self._api.ixn_route_objects[route_range.name] = ixn_bgp_property
        advanced = route_range.get("advanced")
        if (
            advanced is not None
            and advanced.multi_exit_discriminator is not None
        ):
            self.configure_value(
                property_xpath, "enableMultiExitDiscriminator", True
            )
            self.configure_value(
                property_xpath,
                "multiExitDiscriminator",
                advanced.multi_exit_discriminator,
            )
        if advanced is not None:
            self.configure_value(property_xpath, "origin", advanced.origin)
        if route_range.get("as_path") is not None:
            self._config_bgp_as_path(route_range.as_path, ixn_bgp_property)
        if route_range.get("communities"):
            self._config_bgp_community(
                route_range.communities, ixn_bgp_property
            )

    def configure_bgpv6(self, ixn_parent, bgpv6, ixn_dg):
        ixn_bgpv6 = ixn_parent.BgpIpv6Peer
        self._api._remove(ixn_bgpv6, [bgpv6])
        name = self._api.special_char(bgpv6.name)
        args = {
            "Name": name,
        }
        ixn_bgpv6.find(Name="^%s$" % name)
        if len(ixn_bgpv6) == 0:
            ixn_bgpv6.add(**args)[-1]
        else:
            self.update(ixn_bgpv6, **args)
        self._api.ixn_objects[bgpv6.name] = ixn_bgpv6.href
        as_type = "internal"
        if bgpv6.get("as_type") is not None and bgpv6.get("as_type") == "ebgp":
            as_type = "external"
        bgp_xpath = self.get_xpath(ixn_bgpv6.href)
        self.configure_value(bgp_xpath, "type", as_type)
        as_bytes = bgpv6.get("as_number_width")
        if as_bytes is None or as_bytes == "two":
            self.configure_value(bgp_xpath, "localAs2Bytes", bgpv6.as_number)
        elif as_bytes == "four":
            self.configure_value(bgp_xpath, "enable4ByteAs", True)
            self.configure_value(bgp_xpath, "localAs4Bytes", bgpv6.as_number)
        else:
            msg = "Please configure supported [two, four] as_number_width"
            raise Exception(msg)
        if bgpv6.get("dut_address") is not None:
            self.configure_value(bgp_xpath, "dutIp", bgpv6.dut_address)
        if bgpv6.get("as_number_set_mode") is not None:
            self.configure_value(
                bgp_xpath,
                "asSetMode",
                bgpv6.as_number_set_mode,
                enum_map=ConfigureBgp._BGP_AS_SET_MODE,
            )
        # self._configure_pattern(ixn_dg.RouterData.RouterId, bgpv4.router_id)
        if bgpv6.get("advanced") is not None:
            advanced = bgpv6.advanced
            self.configure_value(
                bgp_xpath, "holdTimer", advanced.hold_time_interval
            )
            self.configure_value(
                bgp_xpath, "keepaliveTimer", advanced.keep_alive_interval
            )
            self.configure_value(bgp_xpath, "md5Key", advanced.md5_key)
            self.configure_value(
                bgp_xpath, "updateInterval", advanced.update_interval
            )
            self.configure_value(bgp_xpath, "ttl", advanced.time_to_live)
        if bgpv6.get("sr_te_policies"):
            self._configure_sr_te(ixn_bgpv6, bgp_xpath, bgpv6.sr_te_policies)
        self._bgp_route_builder(ixn_dg, ixn_bgpv6, bgpv6)
        return ixn_bgpv6

    def _configure_bgpv6_route(self, ixn_dg, ixn_bgp, route_range):
        ixn_ng = ixn_dg.NetworkGroup
        name = self._api.special_char(route_range.name)
        args = {
            "Name": name,
        }
        ixn_ng.find(Name="^%s$" % name)
        if len(ixn_ng) == 0:
            self.stop_topology()
            ixn_ng.add(**args)[-1]
            ixn_pool = ixn_ng.Ipv6PrefixPools.add()
        else:
            self.update(ixn_ng, **args)
            ixn_pool = ixn_ng.Ipv6PrefixPools.find()
        ixn_pool.Connector.find().ConnectedTo = ixn_bgp.href
        if route_range.name is not None:
            self._api.ixn_objects[route_range.name] = ixn_ng.href
            self._api._device_encap[route_range.name] = "ipv6"
        pool_infos = self.select_node(
            ixn_pool.href,
            children=["bgpIPRouteProperty", "bgpV6IPRouteProperty"],
        )
        pool_xpath = pool_infos["xpath"]
        addresses = route_range.addresses
        if len(addresses) > 0:
            ixn_ng.Multiplier = len(addresses)
            route_addresses = RouteAddresses()
            for address in addresses:
                route_addresses.address = address.address
                route_addresses.step = address.step
                route_addresses.prefix = address.prefix
                route_addresses.count = address.count
            self.configure_value(
                pool_xpath, "networkAddress", route_addresses.address
            )
            self.configure_value(
                pool_xpath, "prefixAddrStep", route_addresses.step
            )
            self.configure_value(
                pool_xpath, "prefixLength", route_addresses.prefix
            )
            self.configure_value(
                pool_xpath, "numberOfAddressesAsy", route_addresses.count
            )
        if self._api.get_device_encap(ixn_dg.Name) == "ipv4":
            ixn_bgp_property = ixn_pool.BgpIPRouteProperty.find()
            property_xpath = pool_infos["bgpIPRouteProperty"][0]["xpath"]
        else:
            ixn_bgp_property = ixn_pool.BgpV6IPRouteProperty.find()
            property_xpath = pool_infos["bgpV6IPRouteProperty"][0]["xpath"]
        if route_range.get("next_hop_address") is not None:
            self.configure_value(
                property_xpath, "ipv6NextHop", route_range.next_hop_address
            )
        if route_range.name is not None:
            ixn_bgp_property.Name = route_range.name
            self._api.ixn_route_objects[route_range.name] = ixn_bgp_property
        advanced = route_range.get("advanced")
        if (
            advanced is not None
            and advanced.get("multi_exit_discriminator") is not None
        ):
            self.configure_value(
                property_xpath, "enableMultiExitDiscriminator", True
            )
            self.configure_value(
                property_xpath,
                "multiExitDiscriminator",
                advanced.multi_exit_discriminator,
            )
        if advanced is not None:
            self.configure_value(property_xpath, "origin", advanced.origin)
        if route_range.get("as_path") is not None:
            self._config_bgp_as_path(route_range.as_path, ixn_bgp_property)
        if route_range.get("communities") is not None:
            self._config_bgp_community(
                route_range.communities, ixn_bgp_property
            )

    def _config_bgp_as_path(self, as_path, ixn_bgp_property):
        as_path_segments = as_path.as_path_segments
        property_xpath = self.get_xpath(ixn_bgp_property.href)
        if as_path.as_set_mode is not None or len(as_path_segments) > 0:
            self.configure_value(property_xpath, "enableAsPathSegments", True)
            self.configure_value(
                property_xpath,
                "asSetMode",
                as_path.as_set_mode,
                enum_map=ConfigureBgp._BGP_AS_MODE,
            )
            self.configure_value(
                property_xpath,
                "OverridePeerAsSetMode",
                as_path.override_peer_as_set_mode,
            )
            if len(as_path_segments) > 0:
                ixn_bgp_property.NoOfASPathSegmentsPerRouteRange = len(
                    as_path_segments
                )
                ixn_segments = ixn_bgp_property.BgpAsPathSegmentList.find()
                for seg_index, segment in enumerate(as_path_segments):
                    ixn_segment = ixn_segments[seg_index]
                    ixn_segment.SegmentType.Single(
                        ConfigureBgp._BGP_SEG_TYPE[segment.segment_type]
                    )
                    as_numbers = segment.as_numbers
                    if as_numbers is not None:
                        ixn_segment.NumberOfAsNumberInSegment = len(as_numbers)
                        as_numbers_info = self.select_child_node(
                            ixn_segment.href, "bgpAsNumberList"
                        )
                        for as_index, as_number in enumerate(as_numbers):
                            as_num_xpath = as_numbers_info[as_index]["xpath"]
                            self.configure_value(
                                as_num_xpath, "asNumber", as_number
                            )

    def _config_bgp_community(self, communities, ixn_bgp_property):
        if len(communities) == 0:
            ixn_bgp_property.EnableCommunity.Single(False)
            return
        ixn_bgp_property.EnableCommunity.Single(True)
        ixn_bgp_property.NoOfCommunities = len(communities)
        communities_info = self.select_child_node(
            ixn_bgp_property.href, "bgpCommunitiesList"
        )
        for index, community in enumerate(communities):
            community_xpath = communities_info[index]["xpath"]
            if community.community_type is not None:
                self.configure_value(
                    community_xpath,
                    "type",
                    community.community_type,
                    enum_map=ConfigureBgp._BGP_COMMUNITY_TYPE,
                )
            self.configure_value(
                community_xpath, "asNumber", community.as_number
            )
            self.configure_value(
                community_xpath, "lastTwoOctets", community.as_custom
            )

    def _configure_sr_te(self, ixn_bgp, bgp_xpath, sr_te_list):
        if sr_te_list is None or len(sr_te_list) == 0:
            return
        self.configure_value(bgp_xpath, "capabilitySRTEPoliciesV4", True)
        self.configure_value(bgp_xpath, "capabilitySRTEPoliciesV6", True)
        ixn_bgp.NumberSRTEPolicies = len(sr_te_list)
        if re.search("bgpIpv4Peer", ixn_bgp.href) is not None:
            ixn_sr_te = ixn_bgp.BgpSRTEPoliciesListV4
        else:
            ixn_sr_te = ixn_bgp.BgpSRTEPoliciesListV6
        sr_te_xpath = self.get_xpath(ixn_sr_te.href)
        self._configure_attributes(
            ConfigureBgp._BGP_SR_TE, sr_te_list, sr_te_xpath
        )
        next_hops = []
        add_paths = []
        as_paths = []
        communities = []
        for sr_te in sr_te_list:
            if sr_te.get("next_hop") is not None:
                next_hops.append(sr_te.next_hop)
            if sr_te.get("add_path") is not None:
                add_paths.append(sr_te.add_path)
            if sr_te.get("as_path") is not None:
                as_paths.append(sr_te.as_path)
            if sr_te.get("communities") is not None:
                communities.append(sr_te.communities)

        active_list = self._process_nodes(next_hops)
        if active_list != []:
            self.configure_value(sr_te_xpath, "enableNextHop", active_list)
        if any(active_list):
            self._configure_attributes(
                ConfigureBgp._SRTE_NEXT_HOP, next_hops, sr_te_xpath
            )

        active_list = self._process_nodes(add_paths)
        if active_list != []:
            self.configure_value(sr_te_xpath, "enableAddPath", active_list)
        if any(active_list):
            self._configure_attributes(
                ConfigureBgp._SRTE_ADDPATH, add_paths, sr_te_xpath
            )

        active_list = self._process_nodes(as_paths)
        if any(active_list):
            self._configure_attributes(
                ConfigureBgp._SRTE_AS_PATH, as_paths, sr_te_xpath
            )
            self._configure_srte_aspath_segment(as_paths, ixn_sr_te)
        self._configure_tlvs(ixn_sr_te, sr_te_list)

    def _get_symmetric_nodes(self, parent_list, node_name):
        NodesInfo = namedtuple(
            "NodesInfo", ["max_len", "active_list", "symmetric_nodes"]
        )
        nodes_list = []
        max_len = 0
        for parent in parent_list:
            nodes = getattr(parent, node_name)
            node_len = len(nodes)
            if node_len > max_len:
                max_len = node_len
            nodes_list.append(nodes)
        symmetric_nodes = []
        active_list = []
        for nodes in nodes_list:
            if len(nodes) == max_len:
                for node in nodes:
                    active_list.append(node.active)
                    symmetric_nodes.append(node)
            else:
                for index in range(0, max_len):
                    node = nodes[0]
                    if index < len(nodes):
                        node = nodes[index]
                        active_list.append(node.active)
                        symmetric_nodes.append(node)
                    else:
                        active_list.append(False)
                        symmetric_nodes.append(node)
        return NodesInfo(max_len, active_list, symmetric_nodes)

    def _get_symetric_tab_nodes(self, parent_list, node_name):
        TabNodesInfo = namedtuple(
            "TabNodesInfo", ["max_len", "symmetric_nodes_list", "actives_list"]
        )
        max_len = 0
        symmetric_nodes_list = []
        actives_list = []
        is_enable = False
        for parent in parent_list:
            nodes = getattr(parent, node_name)
            if nodes is None:
                continue
            is_enable = True
            node_len = len(nodes)
            if node_len > max_len:
                for index in range(max_len, node_len):
                    symmetric_nodes_list.append(
                        [nodes[index]] * len(parent_list)
                    )
                    actives_list.append([False] * len(parent_list))
                max_len = node_len
        if is_enable:
            for parent_idx, parent in enumerate(parent_list):
                nodes = getattr(parent, node_name)
                for node_idx, node in enumerate(nodes):
                    symmetric_nodes_list[node_idx][parent_idx] = node
                    actives_list[node_idx][parent_idx] = True
        return TabNodesInfo(max_len, symmetric_nodes_list, actives_list)

    def _configure_srte_aspath_segment(self, as_paths, ixn_sr_te):
        nodes_list_info = self._get_symetric_tab_nodes(
            as_paths, "as_path_segments"
        )
        if nodes_list_info.max_len == 0:
            return
        ixn_sr_te.EnableAsPathSegments.Single(True)
        ixn_sr_te.NoOfASPathSegmentsPerRouteRange = nodes_list_info.max_len
        ixn_segments = ixn_sr_te.BgpAsPathSegmentList
        ixn_segments.find()
        segments_info = self.select_node(
            ixn_sr_te.refresh().href, children=["bgpAsPathSegmentList"]
        )
        for seg_idx, segment in enumerate(
            segments_info["bgpAsPathSegmentList"]
        ):
            segment_xpath = segment["xpath"]
            ixn_segment = ixn_segments[seg_idx]
            segment_nodes = nodes_list_info.symmetric_nodes_list[seg_idx]
            self.configure_value(
                segment_xpath,
                "enableASPathSegment",
                nodes_list_info.actives_list[seg_idx],
            )
            self._configure_attributes(
                ConfigureBgp._SRTE_ASPATH_SEGMENT, segment_nodes, segment_xpath
            )
            configure_as_number = False
            for segment_node in segment_nodes:
                as_numbers = getattr(segment_node, "as_numbers")
                if as_numbers is not None:
                    configure_as_number = True
                    if not isinstance(as_numbers, list):
                        raise Exception("as_numbers must be list")
            if configure_as_number is True:
                as_numbers_list = self._get_symetric_tab_nodes(
                    segment_nodes, "as_numbers"
                )
                ixn_segment.NumberOfAsNumberInSegment = as_numbers_list.max_len
                numbers_info = self.select_node(
                    ixn_segment.href, children=["bgpAsNumberList"]
                )
                for num_idx, number in enumerate(
                    numbers_info["bgpAsNumberList"]
                ):
                    number_xpath = number["xpath"]
                    self.configure_value(
                        number_xpath,
                        "enableASNumber",
                        as_numbers_list.actives_list[num_idx],
                    )
                    self.configure_value(
                        number_xpath,
                        "asNumber",
                        as_numbers_list.symmetric_nodes_list[num_idx],
                    )

    def _configure_tlvs(self, ixn_sr_te, sr_te_list):
        nodes_info = self._get_symmetric_nodes(sr_te_list, "tunnel_tlvs")
        if int(nodes_info.max_len) > 2:
            raise Exception(
                "Value {0} for SR TE Policy Number of Tunnel TLVs is "
                "greater than maximal value 2".format(nodes_info.max_len)
            )
        if re.search("bgpSRTEPoliciesListV4", ixn_sr_te.href) is not None:
            ixn_sr_te.NumberOfTunnelsV4 = nodes_info.max_len
            ixn_tunnel = ixn_sr_te.BgpSRTEPoliciesTunnelEncapsulationListV4
        else:
            ixn_sr_te.NumberOfTunnelsV6 = nodes_info.max_len
            ixn_tunnel = ixn_sr_te.BgpSRTEPoliciesTunnelEncapsulationListV6
        tunnel_xpath = self.get_xpath(ixn_tunnel.href)
        self.configure_value(tunnel_xpath, "active", nodes_info.active_list)
        tunnel_tlvs = nodes_info.symmetric_nodes

        remote_endpoint_sub_tlv = []
        preference_sub_tlv = []
        binding_sub_tlv = []
        explicit_null_label_policy_sub_tlv = []
        for tunnel_tlv in tunnel_tlvs:
            if tunnel_tlv.get("remote_endpoint_sub_tlv") is not None:
                remote_endpoint_sub_tlv.append(
                    tunnel_tlv.remote_endpoint_sub_tlv
                )
            if tunnel_tlv.get("preference_sub_tlv") is not None:
                preference_sub_tlv.append(tunnel_tlv.preference_sub_tlv)
            if tunnel_tlv.get("binding_sub_tlv") is not None:
                binding_sub_tlv.append(tunnel_tlv.binding_sub_tlv)
            if (
                tunnel_tlv.get("explicit_null_label_policy_sub_tlv")
                is not None
            ):
                explicit_null_label_policy_sub_tlv.append(
                    tunnel_tlv.explicit_null_label_policy_sub_tlv
                )

        active_list = self._process_nodes(remote_endpoint_sub_tlv)
        if active_list != []:
            self.configure_value(
                tunnel_xpath, "enRemoteEndPointTLV", active_list
            )
        if any(active_list):
            self._configure_attributes(
                ConfigureBgp._REMOTE_ENDPOINT_SUB_TLV,
                remote_endpoint_sub_tlv,
                tunnel_xpath,
            )

        active_list = self._process_nodes(preference_sub_tlv)
        if active_list != []:
            self.configure_value(tunnel_xpath, "enPrefTLV", active_list)
        if any(active_list):
            self._configure_attributes(
                ConfigureBgp._PREFERENCE_SUB_TLV,
                preference_sub_tlv,
                tunnel_xpath,
            )

        active_list = self._process_nodes(binding_sub_tlv)
        if active_list != []:
            self.configure_value(tunnel_xpath, "enBindingTLV", active_list)
        if any(active_list):
            self._configure_attributes(
                ConfigureBgp._BINDING_SUB_TLV, binding_sub_tlv, tunnel_xpath
            )

        active_list = self._process_nodes(explicit_null_label_policy_sub_tlv)
        if active_list != []:
            self.configure_value(tunnel_xpath, "enENLPTLV", active_list)
        if any(active_list):
            self._configure_attributes(
                ConfigureBgp._ENLP_SUB_TLV,
                explicit_null_label_policy_sub_tlv,
                tunnel_xpath,
            )
        self._configure_tlv_segment(ixn_tunnel, tunnel_tlvs)

    def _configure_tlv_segment(self, ixn_tunnel, tunnel_tlvs):
        nodes_info = self._get_symmetric_nodes(tunnel_tlvs, "segment_lists")
        if (
            re.search(
                "bgpSRTEPoliciesTunnelEncapsulationListV4", ixn_tunnel.href
            )
            is not None
        ):
            ixn_tunnel.NumberOfSegmentListV4 = nodes_info.max_len
            ixn_segment_list = ixn_tunnel.BgpSRTEPoliciesSegmentListV4
        else:
            ixn_tunnel.NumberOfSegmentListV6 = nodes_info.max_len
            ixn_segment_list = ixn_tunnel.BgpSRTEPoliciesSegmentListV6
        segment_list_xpath = self.get_xpath(ixn_segment_list.href)
        self.configure_value(
            segment_list_xpath, "active", nodes_info.active_list
        )
        segment_list = nodes_info.symmetric_nodes
        if any(nodes_info.active_list):
            self._configure_attributes(
                ConfigureBgp._POLICIES_SEGMENT_LIST,
                segment_list,
                segment_list_xpath,
            )
            self.configure_value(
                segment_list_xpath, "enWeight", [True] * len(segment_list)
            )
        nodes_info = self._get_symmetric_nodes(segment_list, "segments")
        if (
            re.search("bgpSRTEPoliciesSegmentListV4", ixn_segment_list.href)
            is not None
        ):
            ixn_segment_list.NumberOfSegmentsV4 = nodes_info.max_len
            ixn_segments = ixn_segment_list.BgpSRTEPoliciesSegmentsCollectionV4
        else:
            ixn_segment_list.NumberOfSegmentsV6 = nodes_info.max_len
            ixn_segments = ixn_segment_list.BgpSRTEPoliciesSegmentsCollectionV6
        segments_xpath = self.get_xpath(ixn_segments.href)
        self.configure_value(segments_xpath, "active", nodes_info.active_list)
        segments = nodes_info.symmetric_nodes
        if any(nodes_info.active_list):
            self._configure_attributes(
                ConfigureBgp._SEGMENTS, segments, segments_xpath
            )

    def _process_nodes(self, nodes):
        active_list = []
        for index, node in enumerate(nodes):
            active = False
            if node is None:
                if index == 0:
                    nodes[0] = next(v for v in nodes if v is not None)
                else:
                    nodes[index] = nodes[index - 1]
            else:
                is_config = False
                for name, value in node._properties.items():
                    if value is not None:
                        is_config = True
                        break
                if is_config is True:
                    active = True
            active_list.append(active)
        return active_list

    def _configure_attributes(self, mapper, parent_list, xpath):
        for attribute in mapper:
            attr_mapper = mapper[attribute]
            ixn_attribute = attr_mapper["ixn_attr"]
            default_value = attr_mapper["default"]
            enum_map = attr_mapper.get("enum_map")
            default_obj = getattr(self, str(default_value), None)
            config_values = []
            for parent in parent_list:
                config_value = getattr(parent, attribute, None)
                if config_value is not None:
                    if enum_map is None:
                        config_values.append(str(config_value))
                    else:
                        if str(config_value) not in enum_map.keys():
                            raise Exception(
                                "{0} must configure with enum {1}".format(
                                    attribute, enum_map.keys()
                                )
                            )
                        config_values.append(enum_map[str(config_value)])
                elif default_obj is not None:
                    config_values.append(default_obj())
                else:
                    config_values.append(default_value)
            self.configure_value(xpath, ixn_attribute, config_values)

    def stop_topology(self):
        glob_topo = self._api._globals.Topology.refresh()
        if glob_topo.Status == "started":
            self._api._ixnetwork.StopAllProtocols("sync")


class RouteAddresses(object):
    def __init__(self):
        self._address = []
        self._count = []
        self._prefix = []
        self._step = []

    @property
    def address(self):
        return self._address

    @address.setter
    def address(self, value):
        self._address.append(value)

    @property
    def count(self):
        return self._count

    @count.setter
    def count(self, value):
        self._count.append(value)

    @property
    def prefix(self):
        return self._prefix

    @prefix.setter
    def prefix(self, value):
        self._prefix.append(value)

    @property
    def step(self):
        return self._step

    @step.setter
    def step(self, value):
        self._step.append(value)



from copy import deepcopy


class CustomField(object):
    """
    Implemented All custom field which specify within TrafficItem class
    Best Practice :
        - Please use self._configure_pattern to setting Pattern
    """

    _IPv4_DSCP_PHB = {
        "0": "ipv4.header.priority.ds.phb.defaultPHB.defaultPHB",
        "8": "ipv4.header.priority.ds.phb.classSelectorPHB.classSelectorPHB",
        "16": "ipv4.header.priority.ds.phb.classSelectorPHB.classSelectorPHB",
        "24": "ipv4.header.priority.ds.phb.classSelectorPHB.classSelectorPHB",
        "32": "ipv4.header.priority.ds.phb.classSelectorPHB.classSelectorPHB",
        "40": "ipv4.header.priority.ds.phb.classSelectorPHB.classSelectorPHB",
        "48": "ipv4.header.priority.ds.phb.classSelectorPHB.classSelectorPHB",
        "56": "ipv4.header.priority.ds.phb.classSelectorPHB.classSelectorPHB",
        "10": "ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB",
        "12": "ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB",
        "14": "ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB",
        "18": "ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB",
        "20": "ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB",
        "22": "ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB",
        "26": "ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB",
        "28": "ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB",
        "30": "ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB",
        "34": "ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB",
        "36": "ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB",
        "38": "ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB",
        "46": "ipv4.header.priority.ds.phb.expeditedForwardingPHB.expeditedForwardingPHB",
    }

    _IPv4_DSCP_ECN = {
        "ipv4.header.priority.ds.phb.defaultPHB.defaultPHB": "ipv4.header.priority.ds.phb.defaultPHB.unused",
        "ipv4.header.priority.ds.phb.classSelectorPHB.classSelectorPHB": "ipv4.header.priority.ds.phb.classSelectorPHB.unused",
        "ipv4.header.priority.ds.phb.assuredForwardingPHB.assuredForwardingPHB": "ipv4.header.priority.ds.phb.assuredForwardingPHB.unused",
        "ipv4.header.priority.ds.phb.expeditedForwardingPHB.expeditedForwardingPHB": "ipv4.header.priority.ds.phb.expeditedForwardingPHB.unused",
    }

    _ETHERNETPAUSE = {
        "dst": "ethernet.header.destinationAddress",
        "src": "ethernet.header.sourceAddress",
        "ether_type": "ethernet.header.etherType",
        "control_op_code": "custom.header.length",
        "time": "custom.header.data",
    }

    def _ipv4_priority(self, ixn_field, priority):
        if priority.choice == "dscp":
            field_type_id = None
            first_value = str(self._get_first_value(priority.dscp.phb))
            if first_value in CustomField._IPv4_DSCP_PHB:
                field_type_id = CustomField._IPv4_DSCP_PHB[first_value]
            else:
                field_type_id = CustomField._IPv4_DSCP_PHB["0"]
            if field_type_id is not None:
                self._configure_pattern(
                    ixn_field,
                    field_type_id,
                    priority.dscp.phb,
                    field_choice=True,
                )
                field_type_id = CustomField._IPv4_DSCP_ECN[field_type_id]
                self._configure_pattern(
                    ixn_field,
                    field_type_id,
                    priority.dscp.ecn,
                    field_choice=True,
                )
        elif priority.choice == "tos":
            self._configure_field(ixn_field, priority, field_choice=True)
        else:
            self._configure_pattern(
                ixn_field,
                "ipv4.header.priority.raw",
                priority.raw,
                field_choice=True,
            )

    def _get_first_value(self, phb_pattern):
        if phb_pattern.choice is None:
            return None
        if phb_pattern.choice == "value":
            value = phb_pattern.value
        elif phb_pattern.choice == "values":
            value = phb_pattern.values[0]
        elif phb_pattern.choice == "counter":
            value = phb_pattern.counter.start
        else:
            value = phb_pattern.random.min
        return value

    def adjust_header(self, headers):
        """"""
        new_headers = list()
        for header in headers:
            header = header.parent
            if header.choice == "ethernetpause":
                self._ethernet_pause(
                    new_headers, headers, header.ethernetpause
                )
            elif header.choice == "gtpv1":
                self._gtpv1(new_headers, header)
            else:
                new_headers.append(header)
        return new_headers

    def _ethernet_pause(self, new_headers, headers, ethernetpause):
        """Current IxNetwork 9.10 do not have Global Pause support.
        Therefore using Ether > Custom header as work around.
        We will remove these code and add EthernetPause in generic section.
        The implementation will only support fixed patterns for
        control_op_code and time
        """
        ethpause = deepcopy(ethernetpause)
        headers.clear()
        eth = headers.ethernet()[-1]
        cust = headers.custom()[-1]
        new_headers.append(eth.parent)
        new_headers.append(cust.parent)
        if ethpause._properties.get("src") is not None:
            eth.src.deserialize(ethpause.src.serialize("dict"))
        else:
            eth.src.value = "00:00:aa:00:00:01"
        if ethpause._properties.get("dst") is not None:
            eth.dst.deserialize(ethpause.dst.serialize("dict"))
        else:
            eth.dst.value = "01:80:c2:00:00:01"
        if ethpause._properties.get("ether_type") is not None:
            eth.ether_type.deserialize(ethpause.ether_type.serialize("dict"))
        else:
            eth.ether_type.value = "8808"
        if ethpause._properties.get("control_op_code") is not None:
            control_op_code = self._get_first_value(
                ethpause.control_op_code
            ).zfill(4)
        else:
            control_op_code = "0001"
        if ethpause._properties.get("time") is not None:
            time = self._get_first_value(ethpause.time).zfill(4)
        else:
            time = "FFFF"
        cust.bytes = "{0}{1}".format(control_op_code, time)

    def _gtpv1(self, new_headers, header):
        """"""
        new_headers.append(header)
        gtp_option = deepcopy(header)
        gtp_option.__setattr__("choice", "gtpv1option")
        gtp_option.__setattr__("gtpv1option", header.gtpv1)
        gtp_option.__delattr__("gtpv1")
        new_headers.append(gtp_option)

    def _custom_headers(self, ixn_field, packet):
        if packet.bytes is not None:
            ixn_custom_length = ixn_field.find(
                FieldTypeId="custom.header.length"
            )
            ixn_custom_length.update(
                Auto=False,
                ValueType="singleValue",
                SingleValue=len(packet.bytes) * 4,
            )
            ixn_custom_data = ixn_field.find(FieldTypeId="custom.header.data")
            ixn_custom_data.update(
                Auto=False, ValueType="singleValue", SingleValue=packet.bytes
            )



import re
import time
from collections import namedtuple
from snappi_ixnetwork.exceptions import SnappiIxnException
from snappi_ixnetwork.timer import Timer

try:
    import snappi_convergence
except Exception:
    raise SnappiIxnException(500, "snappi_convergence not installed")
from snappi_ixnetwork.snappi_api import Api as snappiApi


class Api(snappi_convergence.Api):
    _CONVERGENCE = {
        ("data_plane_convergence_us", "DP/DP Convergence Time (us)", float),
        (
            "control_plane_data_plane_convergence_us",
            "CP/DP Convergence Time (us)",
            float,
        ),
    }

    _EVENT = {
        ("begin_timestamp_ns", "Event Start Timestamp", int),
        ("end_timestamp_ns", "Event End Timestamp", int),
    }

    def __init__(self, **kwargs):

        self._convergence_timeout = 10
        self._api = snappiApi(**kwargs)
        self._event_info = None

    def set_config(self, payload):
        try:
            cvg_config = self.convergence_config()
            if isinstance(payload, (type(cvg_config), str)) is False:
                raise TypeError(
                    "The content must be of type Union[Config, str]"
                )

            if isinstance(payload, str) is True:
                payload = cvg_config.deserialize(payload)
            config = payload.get("config")
            if config is None:
                raise Exception("config should not None")
            self._api.config_ixnetwork(config)
            rx_rate_threshold = payload.get("rx_rate_threshold")
            ixn_CpdpConvergence = self._api._traffic.Statistics.CpdpConvergence
            if rx_rate_threshold is not None:
                if self._api.traffic_item.has_latency is True:
                    raise Exception(
                        "We are supporting either latency or rx_rate_threshold"
                    )
                ixn_CpdpConvergence.Enabled = True
                ixn_CpdpConvergence.EnableControlPlaneEvents = True
                ixn_CpdpConvergence.EnableDataPlaneEventsRateMonitor = True
                ixn_CpdpConvergence.DataPlaneThreshold = rx_rate_threshold
            else:
                ixn_CpdpConvergence.Enabled = False
            for ixn_traffic_item in self._api._traffic_item.find():
                ixn_traffic_item.Tracking.find()[0].TrackBy = [
                    "destEndpoint0",
                    "destSessionDescription0",
                ]
        except Exception as err:
            raise SnappiIxnException(err)
        app_errors = self._api._globals.AppErrors.find()
        bad_requests = []
        if len(app_errors) > 0:
            current_errors = app_errors[0].Error.find()
            if len(current_errors) > 0:
                for error in current_errors:
                    if error.Name == "JSON Import Issues":
                        bad_requests = [
                            instance.SourceValues[0]
                            for instance in error.Instance.find()
                        ]
        if bad_requests:
            if len(bad_requests) == 1:
                raise SnappiIxnException(
                    400, "Bad request error: {}".format(bad_requests[0])
                )
            raise SnappiIxnException(400, "Bad request errors:", bad_requests)
        return self._request_detail()

    def set_state(self, payload):
        try:
            cvg_state = self.convergence_state()
            if isinstance(payload, (type(cvg_state), str)) is False:
                raise TypeError(
                    "The content must be of type Union[LinkState, str]"
                )
            if isinstance(payload, str):
                payload = cvg_state.deserialize(payload)
            self._api._connect()
            if payload.choice is None:
                raise Exception("state [transmit/ link/ route] must configure")
            event_names = []
            event_state = None
            event_type = payload.choice
            EventInfo = namedtuple(
                "EventInfo", ["event_type", "event_state", "event_names"]
            )
            if payload.choice == "transmit":
                transmit = payload.transmit
                self._api.traffic_item.transmit(transmit)
            elif payload.choice == "link":
                link = payload.link
                if link.port_names is not None:
                    event_names = link.port_names
                    event_state = link.state
                    self._api.vport.set_link_state(link)
            elif payload.choice == "route":
                route = payload.route
                event_state = route.state
                with Timer(self._api, "Setting route state"):
                    event_names = self._api.ngpf.set_route_state(route)
            else:
                raise Exception(
                    "These[transmit/ link/ route] are valid convergence_state"
                )
            self._event_info = EventInfo(event_type, event_state, event_names)
        except Exception as err:
            raise SnappiIxnException(err)
        return self._request_detail()

    def get_results(self, payload):
        try:
            self._api._connect()
            cvg_req = self.convergence_request()
            if isinstance(payload, (type(cvg_req), str)) is False:
                raise TypeError(
                    "The content must be of type Union[MetricsRequest, str]"
                )
            if isinstance(payload, str) is True:
                payload = cvg_req.deserialize(payload)
            if payload.choice is None:
                raise Exception("state [metrics/ convergence] must configure")
            cvg_res = self.convergence_response()
            if payload.choice == "metrics":
                response = self._api.traffic_item.results(payload.metrics)
                cvg_res.flow_metric.deserialize(response)
            elif payload.choice == "convergence":
                response = self._result(payload.convergence)
                cvg_res.flow_convergence.deserialize(response)
            elif (
                payload.choice
                in self._api.protocol_metrics.get_supported_protocols()
            ):
                response = self._api.protocol_metrics.results(payload)
                getattr(cvg_res, payload.choice + "_metrics").deserialize(
                    response
                )
            else:
                raise Exception(
                    "These[metrics/ convergence] are valid convergence_request"
                )
            return cvg_res
        except Exception as err:
            raise SnappiIxnException(err)

    def _request_detail(self):
        request_detail = self.response_warning()
        errors = self._api._errors
        warnings = list()
        app_errors = self._api._globals.AppErrors.find()
        if len(app_errors) > 0:
            current_errors = app_errors[0].Error.find()
            if len(current_errors) > 0:
                for error in current_errors:
                    # Add loop as sequence are not sorted
                    match = [
                        o
                        for o in self._api._ixn_errors
                        if o.Name == error.Name
                        and o.LastModified == error.LastModified
                    ]
                    if len(match) == 0:
                        if error.ErrorLevel == "kWarning":
                            warnings.append(
                                "IxNet - {0}".format(error.Description)
                            )
                        if error.ErrorLevel == "kError":
                            errors.append(
                                "IxNet - {0}".format(error.Description)
                            )
        # request_detail.errors = errors
        if len(errors) > 0:
            self._api._errors = []
            raise SnappiIxnException(500, errors)
        request_detail.warnings = warnings
        return request_detail

    def _result(self, request):
        flow_names = request.get("flow_names")
        if not isinstance(flow_names, list):
            msg = "Invalid format of flow_names passed {},\
                                                expected list".format(
                flow_names
            )
            raise Exception(msg)
        if flow_names is None or len(flow_names) == 0:
            flow_names = [flow.name for flow in self._api.snappi_config.flows]
        flow_rows = self._get_flow_rows(flow_names)
        traffic_stat = self._api.assistant.StatViewAssistant(
            "Traffic Item Statistics"
        )
        traffic_index = {}
        drill_down_option = "Drill down per Dest Endpoint"
        for index, row in enumerate(
            self._get_traffic_rows(traffic_stat, drill_down_option)
        ):
            traffic_index[row["Traffic Item"]] = index
        response = []
        for flow_name in flow_names:
            convergence = {"name": flow_name}
            if flow_name not in traffic_index.keys():
                raise Exception("Somehow flow %s is missing" % flow_name)

            drill_down_options = traffic_stat.DrillDownOptions()
            drilldown_index = drill_down_options.index(drill_down_option)
            drill_down = traffic_stat.Drilldown(
                traffic_index[flow_name],
                drill_down_option,
                traffic_stat.TargetRowFilters()[drilldown_index],
            )
            drill_down_result = drill_down.Rows[0]
            for (
                external_name,
                internal_name,
                external_type,
            ) in self._CONVERGENCE:
                self._set_result_value(
                    convergence,
                    external_name,
                    drill_down_result[internal_name],
                    external_type,
                )

            events = []
            interruption_time = None
            for flow_result in flow_rows:
                if flow_result["Traffic Item"] != flow_name:
                    continue
                convergence["port_tx"] = flow_result["Tx Port"]
                convergence["port_rx"] = flow_result["Rx Port"]
                event_name = flow_result["Event Name"]
                if event_name == "":
                    continue
                event = self._get_event(event_name, flow_result)
                for external_name, internal_name, external_type in self._EVENT:
                    value = int(flow_result[internal_name].split(".")[-1])
                    self._set_result_value(
                        event, external_name, value * 1000, external_type
                    )
                events.append(event)
                if interruption_time is None:
                    interruption_time = float(
                        flow_result["DP Above Threshold Timestamp"].split(":")[
                            -1
                        ]
                    ) - float(
                        flow_result["DP Below Threshold Timestamp"].split(":")[
                            -1
                        ]
                    )
                    self._set_result_value(
                        convergence,
                        "service_interruption_time_us",
                        interruption_time,
                        float,
                    )
            convergence["events"] = events
            response.append(convergence)
        return response

    def _get_event(self, event_name, flow_result):
        event = {}
        if re.search(r"Port Link Up", event_name):
            if flow_result["Tx Port"] in self._event_info.event_names:
                event["source"] = flow_result["Tx Port"]
            elif flow_result["Rx Port"] in self._event_info.event_names:
                event["source"] = flow_result["Rx Port"]
            else:
                self._api.warning("Not find any event source")
                event["source"] = ""
            if self._event_info.event_state == "up":
                event["type"] = "link_up"
            else:
                event["type"] = "link_down"
        else:
            for route_name in self._api.ixn_route_objects.keys():
                if re.search(route_name, event_name) is not None:
                    event["source"] = route_name
                    event_type = event_name.split(route_name)[-1]
                    if event_type.strip().lower() == "disable":
                        event_type = "route_withdraw"
                    else:
                        event_type = "route_advertise"
                    event["type"] = event_type
                    break
        return event

    def _set_result_value(
        self, row, column_name, column_value, column_type=str
    ):
        try:
            row[column_name] = column_type(column_value)
        except:
            if column_type.__name__ in ["float", "int"]:
                row[column_name] = 0
            else:
                row[column_type] = column_value

    def _get_traffic_rows(self, traffic_stat, drill_down_option):
        count = 0
        sleep_time = 0.5
        while True:
            has_event = False
            drill_down_options = traffic_stat.DrillDownOptions()
            if drill_down_option in drill_down_options:
                has_event = True
                break
            if has_event is True:
                break
            if count * sleep_time > self._convergence_timeout:
                raise Exception(
                    "Somehow 'Drill down per Dest Endpoint' not available"
                )
            time.sleep(sleep_time)
            count += 1
        return traffic_stat.Rows

    def _get_flow_rows(self, flow_names):
        count = 0
        sleep_time = 0.5
        flow_stat = self._api.assistant.StatViewAssistant("Flow Statistics")
        has_flow = False
        while True:
            flow_rows = flow_stat.Rows
            has_event = False
            for row in flow_rows:
                if row["Traffic Item"] in flow_names:
                    has_flow = True
                    if row["Event Name"] != "":
                        has_event = True
                        break
            if has_event is True:
                break
            if count * sleep_time > self._convergence_timeout:
                if has_flow is not True:
                    raise Exception(
                        "flow_names must present within in config.flows"
                    )
                else:
                    raise Exception("Somehow event is not reflected in stat")
            time.sleep(sleep_time)
            count += 1
        return flow_rows



import json
from snappi_ixnetwork.configurebgp import ConfigureBgp


class Ngpf(object):
    """Ngpf configuration

    Args
    ----
    - ixnetworkapi (Api): instance of the ixnetworkapi class
    """

    _TPID_MAP = {
        "8100": "ethertype8100",
        "88a8": "ethertype88a8",
        "9100": "ethertype9100",
        "9200": "ethertype9200",
        "9300": "ethertype9300",
    }

    # Select type of Traffic
    _DEVICE_ENCAP_MAP = {
        "ethernet": "ethernetVlan",
        "ipv4": "ipv4",
        "ipv6": "ipv6",
        "bgpv4": "ipv4",
        "bgpv6": "ipv6",
    }

    _ROUTE_STATE = {"advertise": True, "withdraw": False}

    def __init__(self, ixnetworkapi):
        self._api = ixnetworkapi
        self._conf_bgp = ConfigureBgp(self)

    def config(self):
        """Transform /components/schemas/Device into /topology"""
        self.imports = []
        self._resource_manager = self._api._ixnetwork.ResourceManager
        self._configure_topology(
            self._api._topology, self._api.snappi_config.devices
        )
        self._import(self.imports)

    def update(self, ixn_object, **kwargs):
        update = False
        for name, value in kwargs.items():
            if getattr(ixn_object, name) != value:
                update = True
        if update is True:
            ixn_object.update(**kwargs)

    def _import(self, imports):
        if len(imports) > 0:
            errata = self._resource_manager.ImportConfig(
                json.dumps(imports), False
            )
            for item in errata:
                self._api.warning(item)
            return len(errata) == 0
        return True

    def _configure_topology(self, ixn_topology, devices):
        """One /topology for every unique device.container_name
        Topology name is device.container_name
        """
        topologies = {}
        devices = devices._items
        for device in devices:
            topology = lambda: None
            if device.container_name is None:
                raise NameError("container_name should not None")
            topology.name = self._api._get_topology_name(device.container_name)
            topologies[topology.name] = topology
        self._api._remove(ixn_topology, topologies.values())
        for device in devices:
            name = self._api._get_topology_name(device.container_name)
            ixn_topology.find(Name="^%s$" % self._api.special_char(name))
            if len(ixn_topology) > 0:
                self._api._remove(ixn_topology.DeviceGroup, [device])
        for device in devices:
            args = {
                "Name": self._api._get_topology_name(device.container_name),
                "Ports": [self._api.ixn_objects[device.container_name]],
            }
            ixn_topology.find(
                Name="^%s$" % self._api.special_char(args["Name"])
            )
            if len(ixn_topology) == 0:
                ixn_topology.add(**args)
            else:
                self.update(ixn_topology, **args)
            self._api.ixn_objects[ixn_topology.Name] = ixn_topology.href
            self._configure_device_group(ixn_topology.DeviceGroup, device)

    def _configure_device_group(self, ixn_device_group, device):
        """Transform /components/schemas/Device into /topology/deviceGroup
        One /topology/deviceGroup for every device in port.devices
        """
        args = {"Name": device.name, "Multiplier": 1}
        ixn_device_group.find(
            Name="^%s$" % self._api.special_char(device.name)
        )
        if len(ixn_device_group) == 0:
            ixn_device_group.add(**args)[-1]
        else:
            ixn_ng = ixn_device_group.NetworkGroup
            self._api._remove(ixn_ng, [])
            self.update(ixn_device_group, **args)
        self._config_proto_stack(ixn_device_group, device, ixn_device_group)

    def _config_proto_stack(self, ixn_obj, snappi_obj, ixn_dg):
        self._api.ixn_objects[snappi_obj.name] = ixn_obj.href
        properties = snappi_obj._properties
        for prop_name in properties:
            stack_class = getattr(
                self, "_configure_{0}".format(prop_name), None
            )
            if stack_class is not None:
                child = properties[prop_name]
                if prop_name not in Ngpf._DEVICE_ENCAP_MAP:
                    raise Exception(
                        "Mapping is missing for {0}".format(prop_name)
                    )
                self._api._device_encap[ixn_dg.Name] = Ngpf._DEVICE_ENCAP_MAP[
                    prop_name
                ]
                if child.name is not None:
                    self._api._device_encap[
                        child.name
                    ] = Ngpf._DEVICE_ENCAP_MAP[prop_name]
                new_ixn_obj = stack_class(ixn_obj, child, ixn_dg)
                self._config_proto_stack(new_ixn_obj, child, ixn_dg)

    def _configure_pattern(self, ixn_obj, pattern, enum_map=None):
        if pattern is None:
            return
        # Asymmetric support- without pattern
        if pattern.get("choice") is None:
            if enum_map is not None:
                ixn_obj.Single(enum_map[pattern])
            elif isinstance(pattern, list):
                ixn_obj.ValueList(pattern)
            else:
                ixn_obj.Single(pattern)
        # Symmetric support with pattern
        else:
            if pattern.get("choice") is None:
                return
            elif enum_map is not None and pattern.get("choice") == "value":
                ixn_obj.Single(enum_map[pattern.value])
            elif pattern.get("choice") == "value":
                ixn_obj.Single(pattern.value)
            elif pattern.get("choice") == "values":
                ixn_obj.ValueList(pattern.values)
            elif pattern.get("choice") == "increment":
                ixn_obj.Increment(
                    pattern.increment.start, pattern.increment.step
                )
            elif pattern.get("choice") == "decrement":
                ixn_obj.Decrement(
                    pattern.decrement.start, pattern.decrement.step
                )
            elif pattern.get("choice") == "random":
                pass

    def configure_value(self, source, attribute, value, enum_map=None):
        if value is None:
            return
        xpath = "/multivalue[@source = '{0} {1}']".format(source, attribute)
        if enum_map is not None:
            value = enum_map[value]
        if isinstance(value, list):
            if len(set(value)) == 1:
                ixn_value = {
                    "xpath": "{0}/singleValue".format(xpath),
                    "value": value[0],
                }
            else:
                ixn_value = {
                    "xpath": "{0}/valueList".format(xpath),
                    "values": value,
                }
        else:
            ixn_value = {
                "xpath": "{0}/singleValue".format(xpath),
                "value": value,
            }
        self.imports.append(ixn_value)

    def get_xpath(self, href):
        payload = {
            "selects": [
                {"from": href, "properties": [], "children": [], "inlines": []}
            ]
        }
        url = "%s/operations/select?xpath=true" % self._api._ixnetwork.href
        results = self._api._ixnetwork._connection._execute(url, payload)
        try:
            return results[0]["xpath"]
        except Exception:
            raise Exception("Problem to select %s" % href)

    def select_node(self, href, children=[]):
        payload = {
            "selects": [
                {
                    "from": href,
                    "properties": [],
                    "children": [
                        {
                            "child": "^({0})$".format("|".join(children)),
                            "properties": [],
                            "filters": [],
                        }
                    ],
                    "inlines": [],
                }
            ]
        }
        url = "%s/operations/select?xpath=true" % self._api._ixnetwork.href
        results = self._api._ixnetwork._connection._execute(url, payload)
        try:
            return results[0]
        except Exception:
            raise Exception("Problem to select %s" % href)

    def select_child_node(self, href, child):
        payload = {
            "selects": [
                {
                    "from": href,
                    "properties": [],
                    "children": [
                        {"child": child, "properties": [], "filters": []}
                    ],
                    "inlines": [],
                }
            ]
        }
        url = "%s/operations/select?xpath=true" % self._api._ixnetwork.href
        results = self._api._ixnetwork._connection._execute(url, payload)
        try:
            return results[0][child]
        except Exception:
            raise Exception("Problem to select %s" % href)

    def _configure_ethernet(self, ixn_parent, ethernet, ixn_dg):
        """Transform Device.Ethernet to /topology/.../ethernet"""
        ixn_ethernet = ixn_parent.Ethernet
        self._api._remove(ixn_ethernet, [ethernet])
        args = {}
        ixn_ethernet.find(Name="^%s$" % self._api.special_char(ethernet.name))
        if len(ixn_ethernet) == 0:
            ixn_ethernet.add(**args)
        else:
            self.update(ixn_ethernet, **args)
        if ethernet.name is not None:
            self._api.ixn_objects[ethernet.name] = ixn_ethernet.href
            ixn_ethernet.Name = ethernet.name
        eth_xpath = self.get_xpath(ixn_ethernet.href)
        self.configure_value(eth_xpath, "mac", ethernet.mac)
        self.configure_value(eth_xpath, "mtu", ethernet.mtu)
        if len(ethernet.vlans) > 0:
            ixn_ethernet.VlanCount = len(ethernet.vlans)
            ixn_ethernet.EnableVlans.Single(ixn_ethernet.VlanCount > 0)
            self._configure_vlan(ixn_ethernet.Vlan, ethernet.vlans)
        return ixn_ethernet

    def _configure_vlan(self, ixn_vlans, vlans):
        """Transform Device.Vlan to /topology/.../vlan"""
        for i in range(0, len(ixn_vlans.find())):
            ixn_vlan = ixn_vlans[i]
            if vlans[i].name is not None:
                args = {"Name": vlans[i].name}
                self.update(ixn_vlan, **args)
                self._api.ixn_objects[vlans[i].name] = ixn_vlan.href
            vlan_xpath = self.get_xpath(ixn_vlan.href)
            self.configure_value(vlan_xpath, "vlanId", vlans[i].id)
            self.configure_value(vlan_xpath, "priority", vlans[i].priority)
            self.configure_value(
                vlan_xpath, "tpid", vlans[i].tpid, enum_map=Ngpf._TPID_MAP
            )

    def _configure_ipv4(self, ixn_parent, ipv4, ixn_dg):
        """Transform Device.Ipv4 to /topology/.../ipv4"""
        ixn_ipv4 = ixn_parent.Ipv4
        self._api._remove(ixn_ipv4, [ipv4])
        args = {}
        ixn_ipv4.find(Name="^%s$" % self._api.special_char(ipv4.name))
        if len(ixn_ipv4) == 0:
            ixn_ipv4.add(**args)[-1]
        else:
            self.update(ixn_ipv4, **args)
        if ipv4.name is not None:
            ixn_ipv4.Name = ipv4.name
            self._api.ixn_objects[ipv4.name] = ixn_ipv4.href
        ip_xpath = self.get_xpath(ixn_ipv4.href)
        self.configure_value(ip_xpath, "address", ipv4.address)
        self.configure_value(ip_xpath, "gatewayIp", ipv4.gateway)
        self.configure_value(ip_xpath, "prefix", ipv4.prefix)
        return ixn_ipv4

    def _configure_bgpv4(self, ixn_parent, bgpv4, ixn_dg):
        return self._conf_bgp.configure_bgpv4(ixn_parent, bgpv4, ixn_dg)

    def _configure_ipv6(self, ixn_parent, ipv6, ixn_dg):
        ixn_ipv6 = ixn_parent.Ipv6
        self._api._remove(ixn_ipv6, [ipv6])
        args = {}
        ixn_ipv6.find(Name="^%s$" % self._api.special_char(ipv6.name))
        if len(ixn_ipv6) == 0:
            ixn_ipv6.add(**args)[-1]
        else:
            self.update(ixn_ipv6, **args)
        if ipv6.name is not None:
            ixn_ipv6.Name = ipv6.name
            self._api.ixn_objects[ipv6.name] = ixn_ipv6.href
        ip_xpath = self.get_xpath(ixn_ipv6.href)
        self.configure_value(ip_xpath, "address", ipv6.address)
        self.configure_value(ip_xpath, "gatewayIp", ipv6.gateway)
        self.configure_value(ip_xpath, "prefix", ipv6.prefix)
        return ixn_ipv6

    def _configure_bgpv6(self, ixn_parent, bgpv6, ixn_dg):
        return self._conf_bgp.configure_bgpv6(ixn_parent, bgpv6, ixn_dg)

    def set_route_state(self, payload):
        if payload.state is None:
            return
        names = payload.names
        if len(names) == 0:
            names = self._api.ixn_route_objects.keys()
        for name in names:
            ixn_route = self._api.get_route_object(name)
            ixn_route.Active.Single(Ngpf._ROUTE_STATE[payload.state])
        self._api._ixnetwork.Globals.Topology.ApplyOnTheFly()
        return names



class Events(object):
    def __init__(self, ixnetworkapi):
        self._api = ixnetworkapi

    def config(self):
        events = self._api.snappi_config._properties.get("events")
        if events is None:
            return
        ixn_CpdpConvergence = self._api._traffic.Statistics.CpdpConvergence
        if self.event_allow("rx_rate_threshold"):
            if self._api.traffic_item.has_latency is True:
                raise Exception(
                    "We are supporting either latency or rx_rate_threshold"
                )
            ixn_CpdpConvergence.Enabled = True
            ixn_CpdpConvergence.EnableControlPlaneEvents = True
            ixn_CpdpConvergence.EnableDataPlaneEventsRateMonitor = True
            if events.rx_rate_threshold.threshold is not None:
                ixn_CpdpConvergence.DataPlaneThreshold = (
                    events.rx_rate_threshold.threshold
                )
        else:
            ixn_CpdpConvergence.Enabled = False

    def event_allow(self, event_name):
        events = self._api.snappi_config._properties.get("events")
        if events is None:
            return False
        event = events._properties.get(event_name)
        if events.enable is True or (
            event is not None and event.enable is True
        ):
            return True
        else:
            return False



import yaml
import json


class SerDesObj(object):
    """Comment removed"""

    def __init__(self, data):
        for name, value in data.items():
            setattr(self, name, self._wrap(value))

    def _wrap(self, value):
        if isinstance(value, (tuple, list, set, frozenset)):
            return type(value)([self._wrap(v) for v in value])
        else:
            return SerDesObj(value) if isinstance(value, dict) else value


class SerDes(object):
    def __init__(self):
        pass

    @staticmethod
    def to_object(contents):
        """Given a yaml or json string returns a python object
        If contents is anything else it will just be returned.
        """
        if isinstance(contents, (bytes, str)):
            return SerDes.to_object(yaml.safe_load(contents))
        elif isinstance(contents, dict):
            return SerDesObj(contents)
        else:
            return contents

    @staticmethod
    def to_json(obj):
        """Given a python object returns a json string"""
        return json.dumps(obj, indent=2, default=lambda x: x.__dict__)

    @staticmethod
    def to_yaml(obj):
        """Given a python object returns a yaml string"""
        return yaml.dump(obj)


if __name__ == "__main__":
    # start with yaml
    y = """
config:
  description: |
    A literal description.
    Type-4bits 
     - {some brace} 
     - [some bracket]
  ports:
    - name: port 1
      location: 1.1.1.1/1
    - name: port 2
      location: 1.1.1.1/2
  flows: []
  layer1: []
  captures: []
    """
    # get object from yaml
    o = SerDes.to_object(y)
    # get json from object
    j = SerDes.to_json(o)
    print(j)
    # get object from json
    o = SerDes.to_object(j)
    # get yaml from object
    y = SerDes.to_yaml(o)
    print(y)
